mirror of
https://github.com/ROCm/composable_kernel.git
synced 2026-07-07 15:47:05 +00:00
900 lines
40 KiB
C++
900 lines
40 KiB
C++
// SPDX-License-Identifier: MIT
|
|
// Copyright (c) 2018-2025, Advanced Micro Devices, Inc. All rights reserved.
|
|
|
|
#pragma once
|
|
|
|
#include <iostream>
|
|
#include <string>
|
|
|
|
#include "ck_tile/core.hpp"
|
|
#include "ck_tile/ops/common.hpp"
|
|
#include "ck_tile/host/concat.hpp"
|
|
#include "ck_tile/core/utility/env.hpp"
|
|
|
|
namespace ck_tile {
|
|
|
|
/// @brief The GEMM kernel host arguments.
|
|
///
|
|
/// @par Overview
|
|
/// This structure is passed to @ref GemmKernel "GemmKernel" when creating kernel arguments
|
|
/// object. It contain all necessary information required to build proper kernel argument
|
|
/// and launch kernel on GPU.
|
|
/// This structure defines the GEMM problem configuration by stating all required information
|
|
/// like M,N,K sizes and respective strides.
|
|
/// NumDTensor describes the number of D tensors.
|
|
template <index_t NumATensor = 1, index_t NumBTensor = 1, index_t NumDTensor = 0>
|
|
struct GemmHostArgs
|
|
{
|
|
CK_TILE_HOST GemmHostArgs() = default;
|
|
CK_TILE_HOST GemmHostArgs(const std::array<const void*, NumATensor>& as_ptr_,
|
|
const std::array<const void*, NumBTensor>& bs_ptr_,
|
|
const std::array<const void*, NumDTensor>& ds_ptr_,
|
|
void* e_ptr_,
|
|
index_t k_batch_,
|
|
index_t M_,
|
|
index_t N_,
|
|
index_t K_,
|
|
const std::array<index_t, NumATensor>& stride_As_,
|
|
const std::array<index_t, NumBTensor>& stride_Bs_,
|
|
const std::array<index_t, NumDTensor>& stride_Ds_,
|
|
index_t stride_E_)
|
|
: a_ptr(a_ptr_),
|
|
b_ptr(b_ptr_),
|
|
ds_ptr(ds_ptr_),
|
|
e_ptr(e_ptr_),
|
|
M(M_),
|
|
N(N_),
|
|
K(K_),
|
|
stride_A(stride_A_),
|
|
stride_B(stride_B_),
|
|
stride_Ds(stride_Ds_),
|
|
stride_E(stride_E_),
|
|
k_batch(k_batch_)
|
|
{
|
|
}
|
|
|
|
const std::array<const void*, NumATensor> as_ptr;
|
|
const std::array<const void*, NumBTensor> bs_ptr;
|
|
const std::array<const void*, NumDTensor> ds_ptr;
|
|
void* e_ptr;
|
|
index_t M;
|
|
index_t N;
|
|
index_t K;
|
|
const std::array<index_t, NumATensor> stride_As;
|
|
const std::array<index_t, NumBTensor> stride_Bs;
|
|
const std::array<index_t, NumDTensor> stride_Ds;
|
|
index_t stride_E;
|
|
index_t k_batch;
|
|
};
|
|
|
|
/// @brief The GEMM kernel device arguments.
|
|
template <typename AType = ck_tile::tuple<>, typename BType = ck_tile::tuple<>, typename DType = ck_tile::tuple<>>
|
|
struct GemmKernelArgs
|
|
{
|
|
/// @brief The A input tensor's pointer to device memory.
|
|
const AType* a_ptr;
|
|
/// @brief The B input tensor's pointer to device memory.
|
|
const BType* b_ptr;
|
|
/// @brief The B input tensor's pointer to device memory.
|
|
const DType ds_ptr;
|
|
/// @brief The E output tensor's pointer to device memory.
|
|
void* e_ptr;
|
|
/// @brief GEMM's M dimension size.
|
|
index_t M;
|
|
/// @brief GEMM's N dimension size.
|
|
index_t N;
|
|
/// @brief GEMM's K dimension size.
|
|
index_t K;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of As tensor.
|
|
std::array<index_t, AType::size()> stride_As;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of Bs tensor.
|
|
std::array<index_t, BType::size()> stride_Bs;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of Ds tensor.
|
|
std::array<index_t, DType::size()> stride_Ds;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of E tensor.
|
|
index_t stride_E;
|
|
index_t k_batch;
|
|
};
|
|
|
|
template <index_t NumATensor = 1, index_t NumBTensor = 1>
|
|
struct GemmHostArgs
|
|
{
|
|
CK_TILE_HOST GemmHostArgs() = default;
|
|
CK_TILE_HOST GemmHostArgs(const std::array<const void*, NumATensor>& as_ptr_,
|
|
const std::array<const void*, NumBTensor>& bs_ptr_,
|
|
void* e_ptr_,
|
|
index_t k_batch_,
|
|
index_t M_,
|
|
index_t N_,
|
|
index_t K_,
|
|
const std::array<index_t, NumATensor>& stride_As_,
|
|
const std::array<index_t, NumBTensor>& stride_Bs_,
|
|
index_t stride_E_)
|
|
: as_ptr(as_ptr_),
|
|
bs_ptr(bs_ptr_),
|
|
e_ptr(e_ptr_),
|
|
M(M_),
|
|
N(N_),
|
|
K(K_),
|
|
stride_As(stride_As_),
|
|
stride_Bs(stride_Bs_),
|
|
stride_E(stride_E_),
|
|
k_batch(k_batch_)
|
|
{
|
|
}
|
|
|
|
const std::array<const void*, NumATensor> as_ptr;
|
|
const std::array<const void*, NumBTensor> bs_ptr;
|
|
void* e_ptr;
|
|
index_t M;
|
|
index_t N;
|
|
index_t K;
|
|
const std::array<index_t, NumATensor> stride_As;
|
|
const std::array<index_t, NumBTensor> stride_Bs;
|
|
index_t stride_E;
|
|
index_t k_batch;
|
|
};
|
|
|
|
/// @brief The GEMM kernel device arguments.
|
|
template <typename AType = ck_tile::tuple<>, typename BType = ck_tile::tuple<>>
|
|
struct GemmKernelArgs
|
|
{
|
|
/// @brief The A input tensor's pointer to device memory.
|
|
const AType as_ptr;
|
|
/// @brief The B input tensor's pointer to device memory.
|
|
const BType bs_ptr;
|
|
/// @brief The B input tensor's pointer to device memory.
|
|
void* e_ptr;
|
|
/// @brief GEMM's M dimension size.
|
|
index_t M;
|
|
/// @brief GEMM's N dimension size.
|
|
index_t N;
|
|
/// @brief GEMM's K dimension size.
|
|
index_t K;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of As tensor.
|
|
std::array<index_t, AType::size()> stride_As;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of Bs tensor.
|
|
std::array<index_t, BType::size()> stride_Bs;
|
|
/// @brief The distance between consecutive elements of non-contiguous dimension
|
|
/// (in memory) of E tensor.
|
|
index_t stride_E;
|
|
index_t k_batch;
|
|
};
|
|
|
|
/// @brief The GEMM kernel template.
|
|
///
|
|
/// @paragraph Overview Overview
|
|
/// This class provides the generic matrix multiplication kernel template. By semantic
|
|
/// division of GEMM algorithm into following parts we achieve flexible, versatile
|
|
/// and robust kernel implementation.
|
|
///
|
|
/// @li @b Prolog - The start of GEMM kernel implementation in @ref operator()
|
|
/// function call operator" which determines the work scope of each workgroup.
|
|
/// @li @b GemmPipeline - The core part @a "heart" of matrix multiplication algorithm.
|
|
/// This is the place where each workgroup is loading data from global memory and
|
|
/// carrying out dot products.
|
|
/// @li @b Epilogue - The @a "final" part of matrix multiplication implementation
|
|
/// responsible for storing results to global memory. This is also the place where
|
|
/// any additional operator fusion may take place.
|
|
///
|
|
/// Additionally both @ref GemmPipeline_ "GemmPipeline" and @ref EpiloguePipeline_
|
|
/// "EpiloguePipeline" are parameterized with so called @a Policy which determines all
|
|
/// internal details of those functional parts. You can think of it like both gemm and
|
|
/// epilogue pipelines provides the control-flow logic controlled by policies. Moreover
|
|
/// the policy is responsible for definition of all necessary data layouts and thread's
|
|
/// work distribution.
|
|
///
|
|
/// @tparam TilePartitioner_ The type of class providing mapping of workgroup index into the
|
|
/// output data tile to be calculated. It determines the workgroup to
|
|
/// data relationship (or in other words - which data would be
|
|
/// processed and calculated by which workgroup).
|
|
/// @tparam GemmPipeline_ The type of class which provides the core part of matrix
|
|
/// multiplication. This class should provide implementation of data
|
|
/// loading from global memory and performing block-wise matrix
|
|
/// multiplication. You can think of it as a work done by single
|
|
/// workgroup point of view.
|
|
/// @tparam EpiloguePipeline_ The type of class providing the final part of matrix
|
|
/// multiplication implementation. It is responsible for storing
|
|
/// results calculated by @ref GemmPipeline_ "GemmPipeline" to
|
|
/// the output C tensor in global memory.
|
|
template <typename TilePartitioner_, typename GemmPipeline_, typename EpiloguePipeline_>
|
|
struct GemmKernel
|
|
{
|
|
using TilePartitioner = remove_cvref_t<TilePartitioner_>;
|
|
using GemmPipeline = remove_cvref_t<GemmPipeline_>;
|
|
using EpiloguePipeline = remove_cvref_t<EpiloguePipeline_>;
|
|
using AsLayout = remove_cvref_t<typename GemmPipeline::AsLayout>;
|
|
using BsLayout = remove_cvref_t<typename GemmPipeline::BsLayout>;
|
|
using CLayout = remove_cvref_t<typename GemmPipeline::CLayout>;
|
|
static constexpr index_t KernelBlockSize = GemmPipeline::BlockSize;
|
|
|
|
using AsDataType = remove_cvref_t<typename GemmPipeline::AsDataType>;
|
|
using BsDataType = remove_cvref_t<typename GemmPipeline::BsDataType>;
|
|
// Below type is actually accumulation data type - the output of block GEMM.
|
|
using EDataType = remove_cvref_t<typename EpiloguePipeline::ODataType>;
|
|
|
|
static constexpr auto I0 = number<0>();
|
|
static constexpr auto I1 = number<1>();
|
|
static constexpr auto I2 = number<2>();
|
|
|
|
static constexpr index_t NumATensor = AsDataType::size();
|
|
static constexpr index_t NumBTensor = BsDataType::size();
|
|
|
|
static_assert(AsLayout::size() == AsDataType::size(),
|
|
"The size of AsLayout and AsDataType should be the same");
|
|
|
|
static_assert(BsLayout::size() == BsDataType::size(),
|
|
"The size of BsLayout and BsDataType should be the same");
|
|
|
|
CK_TILE_HOST static constexpr auto MakeAsGridPointer()
|
|
{
|
|
return generate_tuple(
|
|
[&](auto i) {
|
|
using ADataType = remove_cvref_t<std::tuple_element_t<i.value, AsDataType>>;
|
|
|
|
return static_cast<const ADataType*>(nullptr);
|
|
},
|
|
number<NumATensor>{});
|
|
}
|
|
|
|
CK_TILE_HOST static constexpr auto MakeBsGridPointer()
|
|
{
|
|
return generate_tuple(
|
|
[&](auto i) {
|
|
using BDataType = remove_cvref_t<std::tuple_element_t<i.value, BsDataType>>;
|
|
|
|
return static_cast<const BDataType*>(nullptr);
|
|
},
|
|
number<NumBTensor>{});
|
|
}
|
|
|
|
using AsGridPointer = decltype(MakeAsGridPointer());
|
|
using BsGridPointer = decltype(MakeBsGridPointer());
|
|
|
|
[[nodiscard]] CK_TILE_HOST static const std::string GetName()
|
|
{
|
|
// clang-format off
|
|
return concat('_', "gemm", GemmPipeline::GetName());
|
|
// clang-format on
|
|
}
|
|
|
|
CK_TILE_HOST static constexpr auto GridSize(index_t M, index_t N, index_t KBatch)
|
|
{
|
|
return dim3(TilePartitioner::GridSize(M, N), 1, KBatch);
|
|
}
|
|
|
|
CK_TILE_HOST static constexpr auto BlockSize() { return dim3(KernelBlockSize); }
|
|
|
|
CK_TILE_HOST static constexpr GemmKernelArgs<AsGridPointer, BsGridPointer>
|
|
MakeKernelArgs(const GemmHostArgs<NumATensor, NumBTensor>& hostArgs)
|
|
{
|
|
AsGridPointer p_as_grid;
|
|
static_for<0, NumATensor, 1>{}([&](auto index) {
|
|
using ADataType_ = remove_cvref_t<std::tuple_element_t<index.value, AsDataType>>;
|
|
p_as_grid(index) = static_cast<const ADataType_*>(hostArgs.as_ptr[index]);
|
|
});
|
|
|
|
BsGridPointer p_bs_grid;
|
|
static_for<0, NumBTensor, 1>{}([&](auto index) {
|
|
using BDataType_ = remove_cvref_t<std::tuple_element_t<index.value, BsDataType>>;
|
|
p_bs_grid(index) = static_cast<const BDataType_*>(hostArgs.bs_ptr[index]);
|
|
});
|
|
return GemmKernelArgs<AsGridPointer, BsGridPointer>{
|
|
p_as_grid,
|
|
p_bs_grid,
|
|
hostArgs.e_ptr,
|
|
hostArgs.M,
|
|
hostArgs.N,
|
|
hostArgs.K,
|
|
hostArgs.stride_As,
|
|
hostArgs.stride_Bs,
|
|
hostArgs.stride_E,
|
|
hostArgs.k_batch};
|
|
}
|
|
|
|
CK_TILE_HOST_DEVICE static constexpr index_t GetSmemSize()
|
|
{
|
|
return max(GemmPipeline::GetSmemSize(), EpiloguePipeline::GetSmemSize());
|
|
}
|
|
|
|
struct SplitKBatchOffset
|
|
{
|
|
__device__ SplitKBatchOffset(const GemmKernelArgs<AsGridPointer, BsGridPointer>& kargs,
|
|
const std::size_t k_id = blockIdx.z)
|
|
{
|
|
constexpr auto K1 = TilePartitioner::BlockGemmShape::WarpTile::at(number<2>{});
|
|
const index_t K_t = __builtin_amdgcn_readfirstlane(kargs.k_batch * K1);
|
|
const index_t KRead = __builtin_amdgcn_readfirstlane((kargs.K + K_t - 1) / K_t * K1);
|
|
|
|
static_for<0, NumATensor, 1>{}([&](auto index) {
|
|
using AiLayout = remove_cvref_t<std::tuple_element_t<index.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<tensor_layout::gemm::RowMajor, AiLayout>)
|
|
{
|
|
a_k_split_offset[index] = __builtin_amdgcn_readfirstlane(k_id * KRead);
|
|
}
|
|
else if constexpr(std::is_same_v<tensor_layout::gemm::ColumnMajor, AiLayout>)
|
|
{
|
|
a_k_split_offset[index] = __builtin_amdgcn_readfirstlane(k_id * KRead * kargs.stride_As[index]);
|
|
}
|
|
});
|
|
|
|
static_for<0, NumBTensor, 1>{}([&](auto index) {
|
|
using BiLayout = remove_cvref_t<std::tuple_element_t<index.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<tensor_layout::gemm::RowMajor, BiLayout>)
|
|
{
|
|
b_k_split_offset[index] = __builtin_amdgcn_readfirstlane(k_id * KRead * kargs.stride_Bs[index]);
|
|
}
|
|
else if constexpr(std::is_same_v<tensor_layout::gemm::ColumnMajor, BiLayout>)
|
|
{
|
|
b_k_split_offset[index] = __builtin_amdgcn_readfirstlane(k_id * KRead);
|
|
}
|
|
});
|
|
|
|
if(k_id < static_cast<uint32_t>(kargs.k_batch - 1))
|
|
{
|
|
splitted_k = __builtin_amdgcn_readfirstlane(KRead);
|
|
}
|
|
else
|
|
{
|
|
splitted_k = __builtin_amdgcn_readfirstlane(kargs.K - KRead * (kargs.k_batch - 1));
|
|
}
|
|
}
|
|
|
|
std::array<index_t, NumATensor> a_k_split_offset;
|
|
std::array<index_t, NumBTensor> b_k_split_offset;
|
|
index_t splitted_k;
|
|
};
|
|
|
|
CK_TILE_HOST static bool IsSupportedArgument(const GemmKernelArgs<AsGridPointer, BsGridPointer>& kargs)
|
|
{
|
|
if constexpr(EpiloguePipeline::GetVectorSizeC() % 2 != 0 &&
|
|
is_any_of<EDataType, fp16_t, bf16_t>::value)
|
|
{
|
|
if(kargs.k_batch != 1)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("Conditions not met for Kbatch >1 !");
|
|
}
|
|
return false;
|
|
}
|
|
}
|
|
|
|
bool AsTesnorIsValid = {true};
|
|
static_for<0, NumATensor, 1>{}([&](auto index) {
|
|
using AiLayout = remove_cvref_t<std::tuple_element_t<index.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<AiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
if(kargs.K % (TilePartitioner::KPerBlock * kargs.k_batch) != 0 &&
|
|
GemmPipeline::kPadK == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("Can't support K that is not a multiple of k_batch * KPerBlock "
|
|
"without padding!");
|
|
}
|
|
AsTesnorIsValid = false;
|
|
}
|
|
if(kargs.K % GemmPipeline::GetVectorSizeA() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("K is not a multiple of vector load size for A tensor!");
|
|
}
|
|
AsTesnorIsValid = false;
|
|
}
|
|
}
|
|
else
|
|
{
|
|
if(kargs.M % TilePartitioner::MPerBlock != 0 && GemmPipeline::kPadM == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR(
|
|
"Can't support M that is not a multiple of MPerBlock without padding!");
|
|
}
|
|
AsTesnorIsValid = false;
|
|
}
|
|
if(kargs.M % GemmPipeline::GetVectorSizeA() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("M is not a multiple of vector load size for A tensor!");
|
|
}
|
|
AsTesnorIsValid = false;
|
|
}
|
|
}
|
|
});
|
|
|
|
bool BsTesnorIsValid = {true};
|
|
static_for<0, NumBTensor, 1>{}([&](auto index) {
|
|
using BiLayout = remove_cvref_t<std::tuple_element_t<index.value, BsLayout>>;
|
|
if constexpr(std::is_same_v<BiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
if(kargs.N % TilePartitioner::NPerBlock != 0 && GemmPipeline::kPadN == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR(
|
|
"Can't support N that is not a multiple of NPerBlock without padding!");
|
|
}
|
|
BsTesnorIsValid = false;
|
|
}
|
|
if(kargs.N % GemmPipeline::GetVectorSizeB() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("N is not a multiple of vector load size for B tensor!");
|
|
}
|
|
BsTesnorIsValid = false;
|
|
}
|
|
}
|
|
else
|
|
{
|
|
if(kargs.K % (TilePartitioner::KPerBlock * kargs.k_batch) != 0 &&
|
|
GemmPipeline::kPadK == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("Can't support K that is not a multiple of k_batch * KPerBlock "
|
|
"without padding!");
|
|
}
|
|
BsTesnorIsValid = false;
|
|
}
|
|
if(kargs.K % GemmPipeline::GetVectorSizeB() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("K is not a multiple of vector load size for B tensor!");
|
|
}
|
|
BsTesnorIsValid = false;
|
|
}
|
|
}
|
|
});
|
|
|
|
if constexpr(std::is_same_v<CLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
if(kargs.N % TilePartitioner::NPerBlock != 0 && GemmPipeline::kPadN == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR(
|
|
"Can't support N that is not a multiple of NPerBlock without padding!");
|
|
}
|
|
return false;
|
|
}
|
|
if(kargs.N % EpiloguePipeline::GetVectorSizeC() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("N is not a multiple of vector load size for C tensor!");
|
|
}
|
|
return false;
|
|
}
|
|
}
|
|
else
|
|
{
|
|
if(kargs.M % TilePartitioner::MPerBlock != 0 && GemmPipeline::kPadM == false)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR(
|
|
"Can't support M that is not a multiple of MPerBlock without padding!");
|
|
}
|
|
return false;
|
|
}
|
|
if(kargs.M % EpiloguePipeline::GetVectorSizeC() != 0)
|
|
{
|
|
if(ck_tile::EnvIsEnabled(CK_TILE_ENV(CK_TILE_LOGGING)))
|
|
{
|
|
CK_TILE_ERROR("M is not a multiple of vector load size for C tensor!");
|
|
}
|
|
return false;
|
|
}
|
|
}
|
|
return AsTesnorIsValid && BsTesnorIsValid;
|
|
}
|
|
|
|
template <memory_operation_enum DstInMemOp = memory_operation_enum::set>
|
|
CK_TILE_DEVICE static auto MakeGemmTensorViews(const AsGridPointer as_ptr,
|
|
const BsGridPointer bs_ptr,
|
|
EDataType* c_ptr,
|
|
const GemmKernelArgs<AsGridPointer, BsGridPointer>& kargs,
|
|
const SplitKBatchOffset& splitk_batch_offset)
|
|
{
|
|
static_assert(!TilePartitioner::BlockGemmShape::PermuteA, "Not implemented!");
|
|
|
|
const auto& as_tensor_view = generate_tuple([&](auto i) {
|
|
using AiLayout = remove_cvref_t<std::tuple_element_t<i.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<AiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global>(
|
|
as_ptr[i],
|
|
make_tuple(kargs.M, splitk_batch_offset.splitted_k),
|
|
make_tuple(kargs.stride_As[i], 1),
|
|
number<GemmPipeline::GetVectorSizeA()>{},
|
|
number<1>{});
|
|
}
|
|
else
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global>(
|
|
as_ptr[i],
|
|
make_tuple(splitk_batch_offset.splitted_k, kargs.M),
|
|
make_tuple(kargs.stride_As[i], 1),
|
|
number<GemmPipeline::GetVectorSizeA()>{},
|
|
number<1>{});
|
|
}
|
|
}, number<NumATensor>{});
|
|
|
|
const auto& bs_tensor_view = generate_tuple([&](auto i) {
|
|
using BiLayout = remove_cvref_t<std::tuple_element_t<i.value, BsLayout>>;
|
|
if constexpr(std::is_same_v<BiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
if constexpr(TilePartitioner::BlockGemmShape::PermuteB)
|
|
{
|
|
constexpr index_t K1 = GemmPipeline::GetSmemPackB();
|
|
const index_t K0 = splitk_batch_offset.splitted_k / K1;
|
|
constexpr index_t VectorSizeB = std::min(K1, GemmPipeline::GetVectorSizeB());
|
|
const auto b_k0_n_k1_desc =
|
|
make_naive_tensor_descriptor(make_tuple(K0, kargs.N, K1),
|
|
make_tuple(kargs.N * K1, K1, I1),
|
|
number<VectorSizeB>{},
|
|
number<1>{});
|
|
const auto b_n_k_desc = transform_tensor_descriptor(
|
|
b_k0_n_k1_desc,
|
|
make_tuple(make_merge_transform(make_tuple(K0, K1)),
|
|
make_pass_through_transform(kargs.N)),
|
|
make_tuple(sequence<0, 2>{}, sequence<1>{}),
|
|
make_tuple(sequence<0>{}, sequence<1>{}));
|
|
return make_tensor_view<address_space_enum::global>(bs_ptr[i], b_n_k_desc);
|
|
}
|
|
else
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global>(
|
|
bs_ptr,
|
|
make_tuple(splitk_batch_offset.splitted_k, kargs.N),
|
|
make_tuple(kargs.stride_Bs[i], 1),
|
|
number<GemmPipeline::GetVectorSizeB()>{},
|
|
number<1>{});
|
|
}
|
|
}
|
|
else
|
|
{
|
|
if constexpr(TilePartitioner::BlockGemmShape::PermuteB)
|
|
{
|
|
constexpr index_t K1 = GemmPipeline::GetSmemPackB();
|
|
const index_t K0 = splitk_batch_offset.splitted_k / K1;
|
|
constexpr index_t VectorSizeB = std::min(K1, GemmPipeline::GetVectorSizeB());
|
|
const auto b_k0_n_k1_desc =
|
|
make_naive_tensor_descriptor(make_tuple(K0, kargs.N, K1),
|
|
make_tuple(kargs.N * K1, K1, I1),
|
|
number<VectorSizeB>{},
|
|
number<1>{});
|
|
const auto b_n_k_desc = transform_tensor_descriptor(
|
|
b_k0_n_k1_desc,
|
|
make_tuple(make_merge_transform(make_tuple(K0, K1)),
|
|
make_pass_through_transform(kargs.N)),
|
|
make_tuple(sequence<0, 2>{}, sequence<1>{}),
|
|
make_tuple(sequence<1>{}, sequence<0>{}));
|
|
return make_tensor_view<address_space_enum::global>(bs_ptr[i], b_n_k_desc);
|
|
}
|
|
else
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global>(
|
|
bs_ptr[i],
|
|
make_tuple(kargs.N, splitk_batch_offset.splitted_k),
|
|
make_tuple(kargs.stride_Bs[i], 1),
|
|
number<GemmPipeline::GetVectorSizeB()>{},
|
|
number<1>{});
|
|
}
|
|
}
|
|
}, number<NumBTensor>{});
|
|
|
|
// TODO: enable vector write for C in ColMajor
|
|
const auto& c_tensor_view = [&]() {
|
|
if constexpr(std::is_same_v<CLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global, DstInMemOp>(
|
|
c_ptr,
|
|
make_tuple(kargs.M, kargs.N),
|
|
make_tuple(kargs.stride_E, 1),
|
|
number<EpiloguePipeline::GetVectorSizeC()>{},
|
|
number<1>{});
|
|
}
|
|
else
|
|
{
|
|
return make_naive_tensor_view<address_space_enum::global, DstInMemOp>(
|
|
c_ptr,
|
|
make_tuple(kargs.M, kargs.N),
|
|
make_tuple(1, kargs.stride_E),
|
|
number<1>{},
|
|
number<1>{});
|
|
}
|
|
}();
|
|
|
|
return make_tuple(as_tensor_view, bs_tensor_view, c_tensor_view);
|
|
}
|
|
|
|
template <typename TensorView>
|
|
CK_TILE_DEVICE static auto MakeGemmPadViews(const TensorView& views)
|
|
{
|
|
const auto& as_pad_view = generate_tuple(
|
|
[&](auto i) {
|
|
const auto& a_tensor_view = views.at(I0);
|
|
using AiLayout = remove_cvref_t<std::tuple_element_t<i.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<AiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
return pad_tensor_view(a_tensor_view[i],
|
|
make_tuple(number<TilePartitioner::MPerBlock>{},
|
|
number<TilePartitioner::KPerBlock>{}),
|
|
sequence<false, GemmPipeline::kPadK>{});
|
|
}
|
|
else
|
|
{
|
|
return pad_tensor_view(a_tensor_view[i],
|
|
make_tuple(number<TilePartitioner::KPerBlock>{},
|
|
number<TilePartitioner::MPerBlock>{}),
|
|
sequence<false, GemmPipeline::kPadM>{});
|
|
}
|
|
}, number<NumATensor>{});
|
|
|
|
const auto& bs_pad_view = generate_tuple(
|
|
[&](auto i) {
|
|
const auto& b_tensor_view = views.at(I1);
|
|
using BiLayout = remove_cvref_t<std::tuple_element_t<i.value, BsLayout>>;
|
|
if constexpr(std::is_same_v<BiLayout, tensor_layout::gemm::ColumnMajor>)
|
|
{
|
|
return pad_tensor_view(b_tensor_view[i],
|
|
make_tuple(number<TilePartitioner::NPerBlock>{},
|
|
number<TilePartitioner::KPerBlock>{}),
|
|
sequence<false, GemmPipeline::kPadK>{});
|
|
}
|
|
else
|
|
{
|
|
return pad_tensor_view(b_tensor_view[i],
|
|
make_tuple(number<TilePartitioner::KPerBlock>{},
|
|
number<TilePartitioner::NPerBlock>{}),
|
|
sequence<false, GemmPipeline::kPadN>{});
|
|
}
|
|
}, number<NumBTensor>{});
|
|
|
|
// TODO vector write in for C in ColMajor
|
|
const auto& c_pad_view = [&]() {
|
|
const auto& c_tensor_view = views.at(I2);
|
|
if constexpr(std::is_same_v<CLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
return pad_tensor_view(c_tensor_view,
|
|
make_tuple(number<TilePartitioner::MPerBlock>{},
|
|
number<TilePartitioner::NPerBlock>{}),
|
|
sequence<false, GemmPipeline::kPadN>{});
|
|
}
|
|
else
|
|
{
|
|
return pad_tensor_view(c_tensor_view,
|
|
make_tuple(number<TilePartitioner::MPerBlock>{},
|
|
number<TilePartitioner::NPerBlock>{}),
|
|
sequence<GemmPipeline::kPadM, false>{});
|
|
}
|
|
}();
|
|
|
|
return make_tuple(as_pad_view, bs_pad_view, c_pad_view);
|
|
}
|
|
|
|
template <typename PadView>
|
|
CK_TILE_DEVICE static auto
|
|
MakeGemmTileWindows(const PadView& views, const index_t i_m, const index_t i_n)
|
|
{
|
|
const auto& as_pad_view = views.at(I0);
|
|
const auto& bs_pad_view = views.at(I1);
|
|
const auto& c_pad_view = views.at(I2);
|
|
|
|
const auto& as_block_window = generate_tuple(
|
|
[&](auto i) {
|
|
using AiLayout = remove_cvref_t<std::tuple_element_t<i.value, AsLayout>>;
|
|
if constexpr(std::is_same_v<AiLayout, tensor_layout::gemm::RowMajor>)
|
|
{
|
|
return make_tile_window(as_pad_view[i],
|
|
make_tuple(number<TilePartitioner::MPerBlock>{},
|
|
number<TilePartitioner::KPerBlock>{}),
|
|
{i_m, 0});
|
|
}
|
|
else
|
|
{
|
|
return make_tile_window(as_pad_view[i],
|
|
make_tuple(number<TilePartitioner::KPerBlock>{},
|
|
number<TilePartitioner::MPerBlock>{}),
|
|
{0, i_m});
|
|
}
|
|
}, number<NumATensor>{});
|
|
|
|
const auto& bs_block_window = generate_tuple(
|
|
[&](auto i) {
|
|
using BiLayout = remove_cvref_t<std::tuple_element_t<i.value, BsLayout>>;
|
|
if constexpr(std::is_same_v<BiLayout, tensor_layout::gemm::ColumnMajor>)
|
|
{
|
|
return make_tile_window(bs_pad_view[i],
|
|
make_tuple(number<TilePartitioner::NPerBlock>{},
|
|
number<TilePartitioner::KPerBlock>{}),
|
|
{i_n, 0});
|
|
}
|
|
else
|
|
{
|
|
return make_tile_window(bs_pad_view[i],
|
|
make_tuple(number<TilePartitioner::KPerBlock>{},
|
|
number<TilePartitioner::NPerBlock>{}),
|
|
{0, i_n});
|
|
}
|
|
}, number<NumBTensor>{});
|
|
|
|
auto c_block_window = make_tile_window(
|
|
c_pad_view,
|
|
make_tuple(number<TilePartitioner::MPerBlock>{}, number<TilePartitioner::NPerBlock>{}),
|
|
{i_m, i_n});
|
|
|
|
return make_tuple(as_block_window, bs_block_window, c_block_window);
|
|
}
|
|
|
|
/**
|
|
* @brief Runs single GEMM problem cooperatively by whole workgroup.
|
|
*
|
|
* @param as_ptr input As pointer
|
|
* @param bs_ptr input Bs pointer
|
|
* @param c_ptr output C pointer
|
|
* @param smem_ptr_0 The start memory pointer of the shared memory block.
|
|
* @param kargs GEMM kernel arguments
|
|
* @param splitk_batch_offset splitk_batch_offset Utility structure used to calculate k batch.
|
|
* @param block_idx_m The GEMM's output M dimension tile index processed by this workgroup.
|
|
* @param block_idx_n The GEMM's output N dimension tile index processed by this workgroup.
|
|
*
|
|
*/
|
|
CK_TILE_DEVICE static void RunGemm(const AsGridPointer as_ptr,
|
|
const BsGridPointer bs_ptr,
|
|
EDataType* c_ptr,
|
|
void* smem_ptr_0,
|
|
const GemmKernelArgs<AsGridPointer, BsGridPointer>& kargs,
|
|
const SplitKBatchOffset& splitk_batch_offset,
|
|
const index_t block_idx_m,
|
|
const index_t block_idx_n)
|
|
{
|
|
// Create Gemm tensor views, pad views and tile windows
|
|
const auto& gemm_tensor_views_tuple =
|
|
MakeGemmTensorViews<EpiloguePipeline::MemoryOperation>(
|
|
as_ptr, bs_ptr, c_ptr, kargs, splitk_batch_offset);
|
|
|
|
const auto& gemm_pad_views = MakeGemmPadViews(gemm_tensor_views_tuple);
|
|
auto gemm_tile_windows = MakeGemmTileWindows(gemm_pad_views, block_idx_m, block_idx_n);
|
|
|
|
const index_t num_loop = __builtin_amdgcn_readfirstlane(
|
|
TilePartitioner::GetLoopNum(splitk_batch_offset.splitted_k));
|
|
|
|
// Run GEMM cooperatively by whole workgroup.
|
|
auto& as_block_window = gemm_tile_windows.at(I0);
|
|
auto& bs_block_window = gemm_tile_windows.at(I1);
|
|
|
|
//printf("RunGemm before GemmPipeline\n");
|
|
const auto& c_block_tile = GemmPipeline{}.template operator()(
|
|
as_block_window, bs_block_window, num_loop, smem_ptr_0);
|
|
|
|
//printf("RunGemm after GemmPipeline\n");
|
|
// Run Epilogue Pipeline
|
|
auto& c_block_window = gemm_tile_windows.at(I2);
|
|
|
|
EpiloguePipeline{}.template operator()<decltype(c_block_window), decltype(c_block_tile)>(
|
|
c_block_window, c_block_tile, smem_ptr_0);
|
|
}
|
|
|
|
/**
|
|
* @brief Runs single GEMM problem cooperatively by whole workgroup.
|
|
*
|
|
* @note RunGEMM2LDS in with two shared memory buffers using the ping pong buffer mechanism.
|
|
*
|
|
* @param as_ptr input As pointer
|
|
* @param bs_ptr input Bs pointer
|
|
* @param c_ptr output C pointer
|
|
* @param smem_ptr_0 The starting pointer of 1st shared memory block.
|
|
* @param smem_ptr_1 The starting pointer of 2nd shared memory block.
|
|
* @param kargs GEMM kernel arguments
|
|
* @param splitk_batch_offset Utility structure used to calculate k batch.
|
|
* @param block_idx_m The GEMM's output M dimension tile index processed by this workgroup.
|
|
* @param block_idx_n The GEMM's output N dimension tile index processed by this workgroup.
|
|
*
|
|
*/
|
|
CK_TILE_DEVICE static void RunGemm2LDS(const AsGridPointer as_ptr,
|
|
const BsGridPointer bs_ptr,
|
|
EDataType* c_ptr,
|
|
void* __restrict__ smem_ptr_0,
|
|
void* __restrict__ smem_ptr_1,
|
|
const GemmKernelArgs<AsGridPointer, BsGridPointer>& kargs,
|
|
const SplitKBatchOffset& splitk_batch_offset,
|
|
const index_t block_idx_m,
|
|
const index_t block_idx_n)
|
|
{
|
|
// Create Gemm tensor views, pad views and tile windows
|
|
const auto& gemm_tensor_views_tuple =
|
|
MakeGemmTensorViews<EpiloguePipeline::MemoryOperation>(
|
|
as_ptr, bs_ptr, c_ptr, kargs, splitk_batch_offset);
|
|
|
|
const auto& gemm_pad_views = MakeGemmPadViews(gemm_tensor_views_tuple);
|
|
auto gemm_tile_windows = MakeGemmTileWindows(gemm_pad_views, block_idx_m, block_idx_n);
|
|
|
|
const index_t num_loop = __builtin_amdgcn_readfirstlane(
|
|
TilePartitioner::GetLoopNum(splitk_batch_offset.splitted_k));
|
|
|
|
// Run GEMM cooperatively by whole workgroup.
|
|
const auto& as_block_window = gemm_tile_windows.at(I0);
|
|
const auto& bs_block_window = gemm_tile_windows.at(I1);
|
|
|
|
const auto& c_block_tile = GemmPipeline{}.template operator()(
|
|
as_block_window, bs_block_window, num_loop, smem_ptr_0, smem_ptr_1);
|
|
|
|
// Run Epilogue Pipeline
|
|
auto& c_block_window = gemm_tile_windows.at(I2);
|
|
|
|
EpiloguePipeline{}.template operator()<decltype(c_block_window), decltype(c_block_tile)>(
|
|
c_block_window, c_block_tile, smem_ptr_0);
|
|
}
|
|
|
|
CK_TILE_DEVICE void operator()(GemmKernelArgs<AsGridPointer, BsGridPointer> kargs) const
|
|
{
|
|
const auto blockId = __builtin_amdgcn_readfirstlane(blockIdx.x);
|
|
const auto [iM, iN] = TilePartitioner{kargs.M, kargs.N}.GetOutputTileIndex(blockId);
|
|
const index_t i_m = __builtin_amdgcn_readfirstlane(iM * TilePartitioner::MPerBlock);
|
|
const index_t i_n = __builtin_amdgcn_readfirstlane(iN * TilePartitioner::NPerBlock);
|
|
|
|
const SplitKBatchOffset splitk_batch_offset(kargs);
|
|
// options
|
|
AsGridPointer as_ptr;
|
|
static_for<0, NumATensor, 1>{}([&](auto i) {
|
|
using ADataType = remove_cvref_t<std::tuple_element_t<i.value, AsDataType>>;
|
|
as_ptr(i) = static_cast<const ADataType*>(kargs.as_ptr[i]) + splitk_batch_offset.a_k_split_offset[i];
|
|
});
|
|
|
|
BsGridPointer bs_ptr;
|
|
static_for<0, NumBTensor, 1>{}([&](auto i) {
|
|
using BDataType = remove_cvref_t<std::tuple_element_t<i.value, BsDataType>>;
|
|
bs_ptr(i) = static_cast<const BDataType*>(kargs.bs_ptr[i]) + splitk_batch_offset.b_k_split_offset[i];
|
|
});
|
|
|
|
EDataType* c_ptr = static_cast<EDataType*>(kargs.e_ptr);
|
|
|
|
// allocate LDS
|
|
__shared__ char smem_ptr_0[GetSmemSize()];
|
|
|
|
if constexpr(GemmPipeline::DoubleSmemBuffer == true)
|
|
{
|
|
__shared__ char smem_ptr_1[GetSmemSize()];
|
|
if constexpr(!(EpiloguePipeline::MemoryOperation == memory_operation_enum::atomic_add &&
|
|
EpiloguePipeline::GetVectorSizeC() % 2 != 0 &&
|
|
is_any_of<EDataType, fp16_t, bf16_t>::value))
|
|
{
|
|
RunGemm2LDS(as_ptr,
|
|
bs_ptr,
|
|
c_ptr,
|
|
smem_ptr_0,
|
|
smem_ptr_1,
|
|
kargs,
|
|
splitk_batch_offset,
|
|
i_m,
|
|
i_n);
|
|
}
|
|
}
|
|
else
|
|
{
|
|
if constexpr(!(EpiloguePipeline::MemoryOperation == memory_operation_enum::atomic_add &&
|
|
EpiloguePipeline::GetVectorSizeC() % 2 != 0 &&
|
|
is_any_of<EDataType, fp16_t, bf16_t>::value))
|
|
{
|
|
RunGemm(as_ptr, bs_ptr, c_ptr, smem_ptr_0, kargs, splitk_batch_offset, i_m, i_n);
|
|
}
|
|
}
|
|
}
|
|
};
|
|
|
|
} // namespace ck_tile
|